@@ -24,6 +24,7 @@
 #include "nv_include.h"
 #include "nv_rop.h"
 #include "nvc0_accel.h"
+#include "nouveau_copy.h"
 
 #define NOUVEAU_BO(a, b, c) (NOUVEAU_BO_##a | NOUVEAU_BO_##b | NOUVEAU_BO_##c)
 
@@ -289,7 +290,7 @@ NVC0EXACopy(PixmapPtr pdpix, int srcX , int srcY,
 
 	BEGIN_NVC0(push, SUBC_2D(NV50_GRAPH_SERIALIZE), 1);
 	PUSH_DATA (push, 0);
-	BEGIN_NVC0(push, SUBC_2D(0x088c), 1);
+	BEGIN_NVC0(push, NV50_2D(BLIT_CONTROL), 1);
 	PUSH_DATA (push, 0);
 	BEGIN_NVC0(push, NV50_2D(BLIT_DST_X), 12);
 	PUSH_DATA (push, dstX);
@@ -459,14 +460,19 @@ NVC0EXARenderTarget(PixmapPtr ppix, PicturePtr ppict)
 static Bool
 NVC0EXACheckTexture(PicturePtr ppict, PicturePtr pdpict, int op)
 {
-	if (!ppict->pDrawable)
-		NOUVEAU_FALLBACK("Solid and gradient pictures unsupported.\n");
-
-	if (ppict->pDrawable->width > 8192 ||
-	    ppict->pDrawable->height > 8192)
-		NOUVEAU_FALLBACK("texture dimensions exceeded %dx%d\n",
-				 ppict->pDrawable->width,
-				 ppict->pDrawable->height);
+	if (ppict->pDrawable) {
+		if (ppict->pDrawable->width > 8192 ||
+		    ppict->pDrawable->height > 8192)
+			NOUVEAU_FALLBACK("texture too large\n");
+	} else {
+		switch (ppict->pSourcePict->type) {
+		case SourcePictTypeSolidFill:
+			break;
+		default:
+			NOUVEAU_FALLBACK("pict %d\n", ppict->pSourcePict->type);
+			break;
+		}
+	}
 
 	switch (ppict->format) {
 	case PICT_a8r8g8b8:
@@ -659,9 +665,9 @@ NVC0EXAPictTexture(NVPtr pNv, PixmapPtr ppix, PicturePtr ppict, unsigned unit)
 		switch (ppict->repeatType) {
 		case RepeatPad:
 			PUSH_DATA (push, 0x00024000 |
-					 NV50TSC_1_0_WRAPS_CLAMP |
-					 NV50TSC_1_0_WRAPT_CLAMP |
-					 NV50TSC_1_0_WRAPR_CLAMP);
+					 NV50TSC_1_0_WRAPS_CLAMP_TO_EDGE |
+					 NV50TSC_1_0_WRAPT_CLAMP_TO_EDGE |
+					 NV50TSC_1_0_WRAPR_CLAMP_TO_EDGE);
 			break;
 		case RepeatReflect:
 			PUSH_DATA (push, 0x00024000 |
@@ -826,10 +832,8 @@ NVC0EXAPrepareComposite(int op,
 			PicturePtr pspict, PicturePtr pmpict, PicturePtr pdpict,
 			PixmapPtr pspix, PixmapPtr pmpix, PixmapPtr pdpix)
 {
-	struct nouveau_bo *src = nouveau_pixmap_bo(pspix);
 	struct nouveau_bo *dst = nouveau_pixmap_bo(pdpix);
-	struct nouveau_bo *mask = pmpix ? nouveau_pixmap_bo(pmpix) : NULL;
-	NVC0EXA_LOCALS(pspix);
+	NVC0EXA_LOCALS(pdpix);
 
 	if (!PUSH_SPACE(push, 256))
 		NOUVEAU_FALLBACK("space\n");
@@ -881,10 +885,13 @@ NVC0EXAPrepareComposite(int op,
 
 	PUSH_RESET(push);
 	PUSH_REFN (push, pNv->scratch, NOUVEAU_BO_VRAM | NOUVEAU_BO_RDWR);
-	PUSH_REFN (push, src, NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
+	if (pspict->pDrawable)
+		PUSH_REFN (push, nouveau_pixmap_bo(pspix),
+			   NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
 	PUSH_REFN (push, dst, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
-	if (pmpict)
-		PUSH_REFN (push, mask, NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
+	if (pmpict && pmpict->pDrawable)
+		PUSH_REFN (push, nouveau_pixmap_bo(pmpix),
+			   NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
 
 	nouveau_pushbuf_bufctx(push, pNv->bufctx);
 	if (nouveau_pushbuf_validate(push)) {
@@ -1017,65 +1024,6 @@ NVC0EXARectM2MF(NVPtr pNv, int w, int h, int cpp,
 	return TRUE;
 }
 
-Bool
-NVC0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
-		struct nouveau_bo *src, uint32_t src_off, int src_dom,
-		int src_pitch, int src_h, int src_x, int src_y,
-		struct nouveau_bo *dst, uint32_t dst_off, int dst_dom,
-		int dst_pitch, int dst_h, int dst_x, int dst_y)
-{
-	struct nouveau_pushbuf *push = pNv->ce_pushbuf;
-	struct nouveau_pushbuf_refn refs[] = {
-		{ src, src_dom | NOUVEAU_BO_RD },
-		{ dst, dst_dom | NOUVEAU_BO_WR },
-	};
-	unsigned exec;
-
-	if (nouveau_pushbuf_space(push, 64, 0, 0) ||
-	    nouveau_pushbuf_refn (push, refs, 2))
-		return FALSE;
-
-	exec = 0x00000000;
-	if (!src->config.nvc0.memtype) {
-		src_off += src_y * src_pitch + src_x * cpp;
-		exec |= 0x00000010;
-	}
-	if (!dst->config.nvc0.memtype) {
-		dst_off += dst_y * dst_pitch + dst_x * cpp;
-		exec |= 0x00000100;
-	}
-
-	BEGIN_NVC0(push, SUBC_COPY(0x0200), 7);
-	PUSH_DATA (push, src->config.nvc0.tile_mode);
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, src_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, src_x * cpp);
-	PUSH_DATA (push, src_y);
-	BEGIN_NVC0(push, SUBC_COPY(0x0220), 7);
-	PUSH_DATA (push, dst->config.nvc0.tile_mode);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, dst_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, dst_x * cpp);
-	PUSH_DATA (push, dst_y);
-	BEGIN_NVC0(push, SUBC_COPY(0x030c), 8);
-	PUSH_DATA (push, (src->offset + src_off) >> 32);
-	PUSH_DATA (push, (src->offset + src_off));
-	PUSH_DATA (push, (dst->offset + dst_off) >> 32);
-	PUSH_DATA (push, (dst->offset + dst_off));
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, w * cpp);
-	PUSH_DATA (push, h);
-	BEGIN_NVC0(push, SUBC_COPY(0x0300), 1);
-	PUSH_DATA (push, exec);
-
-	return TRUE;
-}
-
 Bool
 NVE0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
 		struct nouveau_bo *src, uint32_t src_off, int src_dom,
@@ -1083,52 +1031,8 @@ NVE0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
 		struct nouveau_bo *dst, uint32_t dst_off, int dst_dom,
 		int dst_pitch, int dst_h, int dst_x, int dst_y)
 {
-	struct nouveau_pushbuf *push = pNv->pushbuf;
-	struct nouveau_pushbuf_refn refs[] = {
-		{ src, src_dom | NOUVEAU_BO_RD },
-		{ dst, dst_dom | NOUVEAU_BO_WR },
-	};
-	unsigned exec;
-
-	if (nouveau_pushbuf_space(push, 64, 0, 0) ||
-	    nouveau_pushbuf_refn (push, refs, 2))
-		return FALSE;
-
-	exec = 0x00000206;
-	if (!src->config.nvc0.memtype) {
-		src_off += src_y * src_pitch + src_x * cpp;
-		exec |= 0x00000080;
-	}
-	if (!dst->config.nvc0.memtype) {
-		dst_off += dst_y * dst_pitch + dst_x * cpp;
-		exec |= 0x00000100;
-	}
-
-	BEGIN_NVC0(push, SUBC_COPY(0x0728), 6);
-	PUSH_DATA (push, 0x00001000 | src->config.nvc0.tile_mode);
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, src_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, (src_y << 16) | src_x * cpp);
-	BEGIN_NVC0(push, SUBC_COPY(0x070c), 6);
-	PUSH_DATA (push, 0x000001000 | dst->config.nvc0.tile_mode);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, dst_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, (dst_y << 16) | dst_x * cpp);
-	BEGIN_NVC0(push, SUBC_COPY(0x0400), 8);
-	PUSH_DATA (push, (src->offset + src_off) >> 32);
-	PUSH_DATA (push, (src->offset + src_off));
-	PUSH_DATA (push, (dst->offset + dst_off) >> 32);
-	PUSH_DATA (push, (dst->offset + dst_off));
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, w * cpp);
-	PUSH_DATA (push, h);
-	BEGIN_NVC0(push, SUBC_COPY(0x0300), 1);
-	PUSH_DATA (push, exec);
-
-	return TRUE;
+	return nouveau_copya0b5_rect(pNv->pushbuf, pNv->NvCOPY, w, h, cpp,
+				     src, src_off, src_dom, src_pitch,
+				     src_h, src_x, src_y, dst, dst_off,
+				     dst_dom, dst_pitch, dst_h, dst_x, dst_y);
 }
